"""
自定义异常
"""


class ForbiddenException(Exception):
    """
    没有权限，可能是token失效
    """


class NotFoundException(Exception):
    """
    无法找到资源
    """


class InternalException(Exception):
    """
    服务器内部错误
    """



"""
nacos for python

"""
from . import func_fuse
from .nacos import Nacos, NacosBalanceClient

__version__ = '1.1.0'

name = 'nacos-starter'

__author__ = 'chizongyang'

__all__ = ['Nacos', 'NacosBalanceClient', 'func_fuse']



"""
nacos module.
"""
import hashlib
import json
import logging
import threading
import time
import typing as t
import urllib
import sys
import requests
import yaml
from requests import Response

from . import util
from .constants import BEAT_TIME, REGISTER_DICT_KEY, TIME_OUT
from .exception import ForbiddenException
from .util import HostPool, logger


def info(msg, *args, **kwargs):
    logger.info("[Nacos] " + msg, *args, **kwargs)

class MediaType:
    """
    http请求参数类型
    """
    MULTIPART_FORM_DATA_VALUE = "multipart/form-data"
    APPLICATION_JSON_VALUE = "application/json"
    APPLICATION_PROBLEM_JSON_VALUE = "application/problem+json"
    APPLICATION_FORM_URLENCODED_VALUE = "application/x-www-form-urlencoded"

    APPLICATION_OCTET_STREAM_VALUE = "application/octet-stream"
    APPLICATION_PDF_VALUE = "application/pdf"
    TEXT_HTML_VALUE = "text/html"
    TEXT_PLAIN_VALUE = "text/plain"
    IMAGE_PNG_VALUE = "image/png"



class Nacos:
    """
    nacos客户端实例，用于注册、获取服务列表等和注册中心、配置中心做交互
    """
    def __init__(self,host="127.0.0.1:8848", username="", password=""):
        # host是ip加端口号，多组用逗号分隔，暂不支持多组
        self.host = host
        self.host_pool = HostPool(host)
        self.username = username
        self.password = password
        self._thread_healthy_dict = {}
        self._config_dict = {}
        self._register_dict = {}
        self.healthy = ""
        self.access_token = "" # token
        self.access_token_invalid_time = -1 # token失效时间
        if username and password:
            login_data = util.get_access_token(
                host=self.host_pool, username=username, password=password)
            if login_data:
                self.access_token = login_data["accessToken"]
                vt = int(time.time()) + login_data["tokenTtl"] - 10 # 设置10秒偏移量
                self.access_token_invalid_time = vt
            else:
                logging.error("nacos认证失败，请检查账号密码是否正确")
                exit(1)
    def __get_host(self):
        """
        获取请求的地址和端口信息，内部通过循环遍历返回一个可用的地址

        Returns
        -------
        nacos客户端
        """
        client = self.host_pool.borrow()
        return client

    def __refresh_token(self):
        """刷新token
        """
        login_data = util.get_access_token(
            host=self.host_pool,
            username=self.username, password=self.password)
        if login_data:
            self.access_token = login_data["accessToken"]
            self.access_token_invalid_time = (int(time.time()) +
                login_data["tokenTtl"] - 10) # 设置10秒偏移量
            return self.access_token
        else:
            logger.error("nacos认证失败，请检查账号密码是否正确")
            exit(1)

    def __get_token(self):
        """检查并获取token
        """
        if self.access_token:
            if time.time() < self.access_token_invalid_time:
                return self.access_token
            else:
                return self.__refresh_token()
        return self.__refresh_token()

    def __wrap_auth_url(self, url = ""):
        """
        包装请求链接，如果需要认证则会拼接accessToken
        """
        token = self.__get_token()
        if token:
            return (url + "&accessToken=" +
                token if "?" in url else url + "?accessToken=" + token)
        return url

    def __healthy_check_thread_run(self):
        """启动健康检查

        register_service注册服务以及config获取配置后，这两个方法会分别启动线程维持心跳,
        本方法用于做异常处理，多次心跳失败后会重新进行注册和配置获取
        """
        while True:
            time.sleep(5)
            self.healthy = int(time.time())
            #检查configThread
            try:
                for item, value in self._config_dict.items():
                    config_msg = item.split("\001")
                    data_id = config_msg[0]
                    group = config_msg[1]
                    tenant = config_msg[2]
                    ht = self._thread_healthy_dict[data_id + group + tenant]
                    x = int(time.time()) - ht
                    if x > 50 and self.config_thread is None:
                        md5_content = config_msg[3]
                        app_config = value
                        self.config_thread = threading.Thread(
                            target=self.__config_listening_thread_run,
                            args=(data_id, group, tenant,
                            md5_content, app_config))
                        hk = data_id + group + tenant
                        self._thread_healthy_dict[hk] = int(time.time())
                        self.config_thread.start()
                        logger.info(
                            "配置信息监听线程重启成功: dataId=%s; group=%s; tenant=%s",
                            data_id, group, tenant)
            except Exception:
                logger.exception("配置信息监听线程健康检查错误",exc_info=True)
            # 检查registerThread
            try:
                x = int(time.time()) - self._register_dict["healthy"]
                if x > 15:
                    service_ip = self._register_dict["serviceIp"]
                    service_port = self._register_dict["servicePort"]
                    service_name = self._register_dict["serviceName"]
                    namespace_id = self._register_dict["namespaceId"]
                    group_name = self._register_dict["groupName"]
                    cluster_name = self._register_dict["clusterName"]
                    ephemeral = self._register_dict["ephemeral"]
                    metadata = self._register_dict["metadata"]
                    weight = self._register_dict["weight"]
                    enabled = self._register_dict["enabled"]
                    self.register_service(service_ip,service_name,service_port,
                                         namespace_id,group_name,cluster_name,
                                         ephemeral,metadata,weight,enabled)
            except Exception:
                logger.exception("服务注册心跳进程健康检查失败",exc_info=True)

    def healthy_check(self):
        """健康检查
        """
        th = threading.Thread(target=self.__healthy_check_thread_run)
        th.start()
        logger.info("健康检查线程已启动")

    def __config_listening_thread_run(self,
        data_id,group,tenant,md5_content,app_config):
        """监听配置修改

        Args:
            dataId: str 服务id
            group: str
            group，默认 DEFAULT_GROUP
            tenant: str 租户
            md5Content: str 内容md5，用于快速比较是否配置有修改
            myConfig: dict 应用当前配置字典
        """
        params = {
            "dataId": data_id,
            "group": group,
            "tenant": tenant
        }

        # 设置长连接30秒，接口会在30秒后返回结果
        header = {"Long-Pulling-Timeout": "30000"}
        dk = data_id + group + tenant
        while True:
            try:
                time.sleep(BEAT_TIME)
            except Exception:
                break
            if not self.config_thread:
                break
            # URL
            get_config_url = self.__wrap_auth_url("http://" +
                self.__get_host().host + "/nacos/v1/cs/configs")
            license_config_url = self.__wrap_auth_url(
                "http://" + self.__get_host().host +
                "/nacos/v1/cs/configs/listener")
            self._thread_healthy_dict[dk] = int(time.time())
            if tenant == "public":
                lck = data_id + "\002" + group + "\002" + md5_content + "\001"
            else:
                lck = "{}\002{}\002{}\002{}\001".format(
                    data_id, group, md5_content, tenant)
            re = requests.post(
                license_config_url,
                data={"Listening-Configs": lck},
                timeout=50,
                headers=header)
            if re.status_code == 403:
                # relogin
                info("获取配置token失效, 准备重新获取")
                self.__refresh_token()
            elif re.status_code == 200:
                if re.text != "":
                    try:
                        # pylint: disable=missing-timeout
                        re = requests.get(get_config_url, params=params)
                        info("获取更新配置内容为\n%s", re.text)
                        nacos_json = self.__get_config_dict(re.text)
                        md5 = hashlib.md5()
                        md5.update(re.content)
                        md5_content = md5.hexdigest()
                        for item in nacos_json:
                            app_config[item] = nacos_json[item]
                        info(
                            "配置信息更新成功: dataId=%s; group=%s; tenant=%s",
                                data_id, group, tenant)
                    except Exception:
                        logger.exception(
                            "配置信息更新失败：dataId=" + data_id + "; group=" +
                            group + "; tenant=" + tenant,
                            exc_info=True)
                        self.config_thread = None
                        break
            else:
                info("获取配置失败终止监听,status_code-%s, message-%s",
                    re.status_code, re.text)
                break
    def __get_data_id(self, env="",file_type="yaml"):
        """获取dataId，用于定位到nacos配置文件
        """
        if env:
            env = "-" + env
        return "{}{}.{}".format(self._register_dict["serviceName"],
            env, file_type)

    def __get_config_dict(self, content: str, file_type="yaml"):
        """获取配置信息转换为字典

        Args:
          content: 内容
          type: 配置类型
        """
        try:
            if file_type == "yaml":
                return yaml.load(content, Loader=yaml.Loader)
            if file_type == "json":
                return json.loads(content)
        except Exception:
            return {}

    def config(self,app_config,env="",file_type="yaml",
        group="DEFAULT_GROUP",tenant="public"):
        """开始执行配置读取

        Args:
          app_config: 应用配置字典
          env: 环境，用于拼接data_id，生成格式为：SERVICE_NAME-env.file_type
          file_type: 文件类型，对应nacos可配置的文件类型，例如json、text、yaml等，默认yaml
          group: group，和nacos中配置对应，默认 DEFAULT_GROUP
          tenant: 租户，和nacos中配置对应，默认 public
        """
        data_id = self.__get_data_id(env=env, file_type=file_type)
        logger.info("正在获取配置: dataId="+
            data_id+"; group="+group+"; tenant="+tenant)
        get_config_url = self.__wrap_auth_url(
            "http://" + self.__get_host().host + "/nacos/v1/cs/configs")
        params = {
            "dataId": data_id,
            "group": group,
            "tenant": tenant
        }
        try:
            # pylint: disable=missing-timeout
            re = requests.get(get_config_url, params=params)
            if re.status_code != 200:
                logger.warning("配置获取失败：dataId="+
                    data_id+"; group="+group+"; tenant="+tenant)
                return
            logging.info("[Nacos] config: %s", re.text)
            nacos_json = self.__get_config_dict(re.text, file_type=file_type)
            md5 = hashlib.md5()
            md5.update(re.content)
            md5_content = md5.hexdigest()
            dk = data_id+"\001"+group+"\001"+tenant+"\001"+md5_content
            self._config_dict[dk] = app_config

            for item in nacos_json:
                app_config[item] = nacos_json[item]
            logger.info("配置获取成功：dataId=%s; group=%s; tenant=%s",
                data_id, group, tenant)
            self.config_thread = threading.Thread(
                target=self.__config_listening_thread_run,
                args=(data_id,group,tenant,md5_content,app_config))
            self._thread_healthy_dict[data_id+group+tenant] = int(time.time())
            self.config_thread.start()
        except Exception:
            logger.exception("配置获取失败：dataId="+
                data_id+"; group="+group+"; tenant="+tenant, exc_info=True)

    def __register_beat_thread_run(self,service_ip,service_port,service_name,
                                group_name,namespace_id,metadata,weight):
        """"
        注册心跳检测
        """
        beat_json = {
            "ip": service_ip,
            "port": service_port,
            "serviceName": service_name,
            "metadata": metadata,
#            "scheduled": "true",
            "weight": weight
        }
        params_beat = {
            "serviceName": service_name,
            "groupName": group_name,
            "namespaceId": namespace_id,
            "beat": urllib.request.quote(json.dumps(beat_json))
        }

        while True:
            self._register_dict[REGISTER_DICT_KEY] = int(time.time())
            try:
                time.sleep(BEAT_TIME)
                re = self.__get_host().beat(
                    access_token=self.access_token, params=params_beat)
                if (re is None or re.status_code != 200
                    or re.json()["code"] != 10200):
                    self._register_dict[REGISTER_DICT_KEY] = int(time.time())-10
                    logger.warning("[Nacos] 心跳请求失败: %s", (re and re.text))

                if re is not None and re.status_code == 403:
                    self.__refresh_token()
                    logger.info("[Nacos] 重新刷新token结果: %s", self.access_token)
            except json.JSONDecodeError:
                self._register_dict[REGISTER_DICT_KEY] = int(time.time()) - 10
                break
            except Exception:
                logger.exception("服务心跳维持失败！", exc_info=True)
                break

    def register_service(self,service_ip,
                        service_name,service_port=80,namespace_id="public",
                        group_name="DEFAULT_GROUP",cluster_name="DEFAULT",
                        ephemeral=True,metadata=None,weight=1,enabled=True):
        """注册服务

        将当前服务注册到nacos

        Args:
          service_ip: 当前服务的ip，会注册到nacos被其他服务来调用，如果不设置会自动获取当前服务器的IP
          service_name: 服务名称
          service_port: 当前服务的端口号，用于被其他服务调用
          namespace_id: 命名空间，默认 public
          group_name: 注册的group 默认 DEFAULT_GROUP
          cluster_name: nacos集群名称，默认 DEFAULT
          ephemeral: 默认True
          metadata: 注册到nacos时携带的元数据
          weight: 权重 默认1
          enabled: 是否启用 默认True
        """
        service_ip = service_ip or util.get_host_ip()
        self._register_dict["serviceIp"] = service_ip
        self._register_dict["servicePort"] = service_port
        self._register_dict["serviceName"] = service_name
        self._register_dict["namespaceId"] = namespace_id
        self._register_dict["groupName"] = group_name
        self._register_dict["clusterName"] = cluster_name
        self._register_dict["ephemeral"] = ephemeral
        self._register_dict["metadata"] = metadata or {}
        self._register_dict["weight"] = weight
        self._register_dict["enabled"] = enabled

        self._register_dict["healthy"] = int(time.time())

        params = {
            "ip": service_ip,
            "port": service_port,
            "serviceName": service_name,
            "namespaceId": namespace_id,
            "groupName": group_name,
            "clusterName": cluster_name,
            "ephemeral": ephemeral,
            "metadata": json.dumps(metadata),
            "weight": weight,
            "enabled": enabled
        }
        try:
            re = self.__get_host().regist_service(
                access_token=self.access_token, params=params)
            if re == "ok":
                logger.info("服务注册成功。")
                beat_thread = threading.Thread(
                    target=self.__register_beat_thread_run,
                    args=(service_ip,service_port,service_name,
                    group_name,namespace_id,metadata,weight))
                beat_thread.start()
            else:
                logger.error("服务注册失败 %s", re)
        except ForbiddenException:
            self.__refresh_token()
        except Exception:
            logger.exception("服务注册失败",exc_info=True)

def default_fallback_fun():
    return "request Error"
def default_time_out_fun():
    return "request time out"





class NacosBalanceClient:
    """
    用于调用其他服务接口, 类似java里边的feign, 接口调用支持负载均衡
    """
    def __init__(self,host="127.0.0.1:8848",service_name="", username="",
        password="",
        group="DEFAULT_GROUP",namespace_id="public",timeout=TIME_OUT,
        fallback_fun=default_fallback_fun, time_out_fun=default_time_out_fun):
        """_summary_

        Args:
            host (str, optional): nacos地址. Defaults to "127.0.0.1:8848".
            service_name (str, optional): 要调用的服务名，在nacos管理页面查看. Defaults to "".
            username (str, optional): nacos账号. Defaults to "".
            password (str, optional): nacos密码. Defaults to "".
            group (str, optional): group. Defaults to "DEFAULT_GROUP".
            namespace_id (str, optional): namespace的id. Defaults to "public".
            timeout (_type_, optional): 超时时间, 单位秒. Defaults to 10.
            fallback_fun (_type_, optional): 失败回调. Defaults to default_fallback_fun.
            time_out_fun (_type_, optional): 超时回调. Defaults to default_time_out_fun.
        """
        self.host = host
        self.host_pool = HostPool(host)
        self.service_name = service_name
        self.group = group
        self.namespace_id = namespace_id
        self._load_balance_dict = {}
        self.timeout = timeout or TIME_OUT
        self.fallback_fun = fallback_fun
        self.time_out_fun  = time_out_fun
        self.username = username
        self.password = password
        self.access_token = "" # token
        self.access_token_invalid_time = -1 # token失效时间
        if username and password:
            login_data = util.get_access_token(host=self.host_pool,
                username=username, password=password)
            if login_data:
                self.access_token = login_data["accessToken"]
                it = login_data["tokenTtl"] - 10 # 设置10秒偏移量
                self.access_token_invalid_time = int(time.time()) + it
            else:
                logger.error("nacos认证失败，请检查账号密码是否正确")
                sys.exit(1)
    def __get_host(self):
        """
        获取请求的地址和端口信息，内部通过循环遍历返回一个可用的地址
        """
        client = self.host_pool.borrow()
        return client

    def __refresh_token(self):
        """刷新token
        """
        login_data = util.get_access_token(
            host=self.host_pool,
            username=self.username, password=self.password)
        if login_data:
            self.access_token = login_data["accessToken"]
            self.access_token_invalid_time = (int(time.time()) +
                login_data["tokenTtl"] - 10) # 设置10秒偏移量
            return self.access_token
        else:
            logger.error("nacos认证失败，请检查账号密码是否正确")
            sys.exit(1)

    def __get_token(self):
        """检查并获取token
        """
        if self.access_token:
            if time.time() < self.access_token_invalid_time:
                return self.access_token
            else:
                return self.__refresh_token()
        return self.__refresh_token()


    def __wrap_auth_url(self, url = ""):
        """
        包装请求链接，如果需要认证则会拼接accessToken
        """
        token = self.__get_token()

        if token:
            return (url + "&accessToken=" +
                token if "?" in url else url + "?accessToken=" + token)
        return url

    def __do_request(self,
        method,url,request_param_json,
        *args,
        produces=MediaType.TEXT_PLAIN_VALUE,
        **kwargs) :
        method = str.upper(method or "GET")
        resp: Response
        if method == "GET":
            url = url + "/"
            for item in args:
                url = url + str(item) + "/"
            url = url[:-1]
            if len(kwargs) != 0:
                url = url + "?"
                for item, value in kwargs.items():
                    url = url + str(item) + "=" + str(value) + "&"
                url = url[:-1]
            logger.debug("feign请求接口 %s", url)
            resp = requests.get(url, timeout=self.timeout)
        if method == "POST":
            if request_param_json:
                header = {"Content-type": "application/json;charset=utf-8"}
                data = None
                for item in args:
                    data = item
                resp = requests.post(url,headers=header,
                    data=json.dumps(data,ensure_ascii=False).encode("utf-8"),
                    timeout=self.timeout)
            else:
                files = {}
                for arg in args:
                    for key in arg:
                        files[key] = (None,arg[key])
                resp = requests.post(url,files=files, timeout=self.timeout)
        if method == "PUT":
            if request_param_json:
                header = {"Content-type": "application/json;charset=utf-8"}
                data = None
                for item in args:
                    data = item
                resp = requests.put(url,headers=header,
                    data=json.dumps(data,ensure_ascii=False).encode("utf-8"),
                    timeout=self.timeout)
            else:
                files = {}
                for arg in args:
                    for key in arg:
                        files[key] = (None,arg[key])
                resp = requests.put(url,files=files, timeout=self.timeout)
        if resp:
            if resp.status_code == 403:
                self.__refresh_token()
                return "Token forbidden, refreshed now"
            if (not produces or produces == MediaType.TEXT_PLAIN_VALUE
                or produces == MediaType.TEXT_HTML_VALUE):
                return resp.text
            if produces == MediaType.APPLICATION_OCTET_STREAM_VALUE:
                return resp.content
            if produces == MediaType.APPLICATION_JSON_VALUE:
                return resp.json()
        return "Request Error"


    def __get_address(self,service_name,group,namespace_id):
        get_provider_url = ("http://" + self.__get_host().host +
            "/nacos/v1/ns/instance/list")
        params = {
            "serviceName": service_name,
            "groupName": group,
            "namespaceId": namespace_id
        }
        get_provider_url = self.__wrap_auth_url(get_provider_url)
        # pylint: disable=missing-timeout
        re = requests.get(get_provider_url, params=params)
        if re.status_code != 200:
            logger.error("nacos返回状态码[%s], 信息：%s", re.status_code, re.text)
            sys.exit(0)
        try:
            msg = re.json()["hosts"]
        except json.JSONDecodeError:
            msg = []
        hosts = []
        for item in msg:
            hosts.append({
                "ip": item["ip"],
                "port": item["port"],
                "healthy": item["healthy"]
            })
        md5 = hashlib.md5()
        md5.update(json.dumps(hosts,ensure_ascii=False).encode("utf-8"))
        md5_content = md5.hexdigest()
        try:
            old_md5 = self._load_balance_dict[
                service_name + group + namespace_id + "md5"]
        except KeyError:
            self._load_balance_dict[
                service_name + group + namespace_id + "md5"] = md5_content
            old_md5 = ""
        if old_md5 != md5_content:
            healthy_hosts = []
            for host in msg:
                if host["healthy"] is True:
                    healthy_hosts.append(host)
            self._load_balance_dict[
                service_name + group + namespace_id] = healthy_hosts
            self._load_balance_dict[
                service_name + group + namespace_id + "index"] = 0

    def __load_balance_client(self,service_name,group,namespace_id):
        try:
            x = int(time.time()) - self._load_balance_dict[
                service_name + group + namespace_id + "time"]
        except KeyError:
            x = 11
        if x > 10:
            self.__get_address(service_name,group,namespace_id)
            self._load_balance_dict[
                service_name + group + namespace_id + "time"] = int(time.time())

        index = self._load_balance_dict[
            service_name + group + namespace_id + "index"]
        l = len(self._load_balance_dict[service_name + group + namespace_id])
        if l == 0:
            logger.error("无可用服务 serviceName: %s;group: %s;namespaceId: %s",
                service_name, group, namespace_id)
            return ""
        if index >= l:
            ip = str(self._load_balance_dict[
                service_name + group + namespace_id][0]["port"])
            self._load_balance_dict[
                service_name + group + namespace_id + "index"] = 1
            return self._load_balance_dict[
                service_name + group + namespace_id][0]["ip"]+":"+ip
        else:
            self._load_balance_dict[
                service_name + group + namespace_id + "index"] = index + 1
            ip = str(self._load_balance_dict[
                service_name + group + namespace_id][index]["port"])
            return  self._load_balance_dict[
                service_name + group + namespace_id][index]["ip"] + ":" + ip

    def request_mapping(self,method,url,
                            request_param_json=False,https=False,
                            consumers=MediaType.APPLICATION_JSON_VALUE,
                            produces=MediaType.TEXT_PLAIN_VALUE) -> t.Callable:
        """
        接口映射配置，同java feign，默认请求参数为json格式，返回结果为文本
        """
        def decorator(f: t.Callable):
            def main_pro(*args, **kwargs):
                address = self.__load_balance_client(
                    self.service_name, self.group, self.namespace_id)
                if address == "":
                    return
                else:
                    if https:
                        request_url = "https://" + address + url
                    else:
                        request_url = "http://" + address + url
                    try:
                        return self.__do_request(
                            method, request_url, request_param_json,
                            consumers=consumers, produces=produces,
                            *args, **kwargs)
                    except requests.ConnectTimeout:
                        logger.exception("链接超时   ",exc_info=True)
                        return self.time_out_fun(self.service_name,
                            self.group,self.namespace_id,method,url)
                    except Exception as ex:
                        logger.exception("链接失败   ", exc_info=True)
                        return self.fallback_fun(self.service_name,
                            self.group,self.namespace_id,method,url,ex)
            main_pro.__name__ = f.__name__
            return main_pro
        return decorator

    def put_mapping(self, **kwargs) -> t.Callable:
        return self.request_mapping(method="PUT", **kwargs)

    def post_mapping(self, **kwargs) -> t.Callable:
        return self.request_mapping(method="POST", **kwargs)

    def get_mapping(self, **kwargs) -> t.Callable:
        return self.request_mapping(method="GET", **kwargs)



"""
nacos operate foundation utils.
"""
import socket
import requests
from requests.exceptions import ConnectionError as RequestConnectionError
import logging
import time as dt
from requests import Response
from typing import List, Any
from .constants import LOG_FORMAT, DATE_FORMAT
from .exception import ForbiddenException, NotFoundException, InternalException

#日志配置
logging.basicConfig(level=logging.INFO,format=LOG_FORMAT,datefmt=DATE_FORMAT)
logger = logging.getLogger("flask.app")

if not logger.hasHandlers():
    logger = logging

# 默认请求超时时间 5秒
DEFAULT_TIMEOUT = 5

class MediaType:
    """
    http请求参数类型
    """
    MULTIPART_FORM_DATA_VALUE = "multipart/form-data"
    APPLICATION_JSON_VALUE = "application/json"
    APPLICATION_PROBLEM_JSON_VALUE = "application/problem+json"
    APPLICATION_FORM_URLENCODED_VALUE = "application/x-www-form-urlencoded"

    APPLICATION_OCTET_STREAM_VALUE = "application/octet-stream"
    APPLICATION_PDF_VALUE = "application/pdf"
    TEXT_HTML_VALUE = "text/html"
    TEXT_PLAIN_VALUE = "text/plain"
    IMAGE_PNG_VALUE = "image/png"
    # 原始response对象
    ORIGIN_RESPONSE = "response"


def do_request(method,url, *args,
    response_type=MediaType.APPLICATION_JSON_VALUE, **kwargs) :
    method = str.upper(method or "GET")
    resp: Response
    logger.debug("[nacos] 请求 - [%s] - %s", method, url)
    if "timeout" not in kwargs:
        kwargs["timeout"] = DEFAULT_TIMEOUT
    if method == "GET":
        url = url + "/"
        for item in args:
            url = url + str(item) + "/"
        url = url[:-1]
        if len(kwargs) != 0:
            url = url + "?"
            for item, value in kwargs.items():
                url = url + str(item) + "=" + str(value) + "&"
            url = url[:-1]
        logger.debug("请求接口 %s", url)
        # pylint: disable=missing-timeout
        resp = requests.get(url, *args, **kwargs)
    if method == "POST":
        # pylint: disable=missing-timeout
        resp = requests.post(url, *args, **kwargs)
    if method == "PUT":
        # pylint: disable=missing-timeout
        return requests.put(url, *args, **kwargs)
    if resp is None:
        return "Request Error"
    elif resp:
        try:
            if MediaType.ORIGIN_RESPONSE == response_type:
                return resp
            if (not response_type or response_type == MediaType.TEXT_PLAIN_VALUE
                or response_type == MediaType.TEXT_HTML_VALUE):
                return resp.text
            if response_type == MediaType.APPLICATION_OCTET_STREAM_VALUE:
                return resp.content
            if response_type == MediaType.APPLICATION_JSON_VALUE:
                return resp.json()
        finally:
            resp.close()
    else:
        if resp.status_code == 403:
            raise ForbiddenException()
        if resp.status_code == 404:
            raise NotFoundException()
        if resp.status_code == 500:
            raise InternalException()

class HostClient:
    """
    nacos连接客户端，提供基本的操作方法，每个客户端对应一个host地址
    """
    host = ""
    protocol = "http"
    # 失败尝试次数
    trys = 5
    is_valid = True
    # 是否健康，上次请求成功则标记健康
    is_health = True

    def __init__(self, host="127.0.0.1:8848", protocol="http", trys=5):
        self.host = host
        self.trys = trys
        self.protocol = protocol

    def __get_prefix_uri(self):
        return f"{self.protocol}://{self.host}/nacos/v1"

    def __request(self, uri: str, method: str, *args,
        data=None, response_type=MediaType.APPLICATION_JSON_VALUE,
        **kwargs):
        """
        执行接口请求

        Args:
          uri: url地址
          method: url请求方法，e.g. get
          data: body参数
          response_type: 返回结果格式，用于对返回结果自动处理，默认json

        Returns:
          根据返回结果类型处理过的接口调用结果，或者请求失败返回None
        """
        url = self.__get_prefix_uri() + uri
        try:
            result = do_request(method=method, url=url,
                data=data, response_type=response_type, *args, **kwargs)
            self.is_health = True
            return result
        except RequestConnectionError as e:
            self.trys = self.trys - 1
            logger.warning("[Nacos] URL [%s] 请求失败：%s", url, str(e))
            self.is_health = False


    def login(self, username: str, password: str) -> Any:
        """
        登录

        Args:
          username: 用户名
          password: 密码

        Returns:
          token数据，如果登陆成功
          None，如果失败

        Examples:
          client.login("nacos", "nacos")
        """
        data = self.__request("/auth/login", "post", data={
            "username": username,
            "password": password
        })
        return data

    def regist_service(self, access_token="", params=None):
        """"
        注册服务

        Parameters
        ----------
        access_token : str
            token，通过登录获取
        params : dict
            注册服务用到的参数

        Returns
        -------
        注册结果
        """
        return (self.__request("/ns/instance?accessToken=" +
            access_token, "post", response_type=MediaType.TEXT_PLAIN_VALUE,
            params=params or {}) or "请求失败")

    def beat(self, access_token="", params=None) -> Response:
        """
        维持心跳

        Args:
            access_token: token，通过登录获取
            params: 用于维持心跳的请求参数

        Returns:
            response对象
        """
        uri = "/ns/instance/beat?accessToken=" + access_token
        if params:
            for item in params:
                uri = uri + f"&{item}={params[item]}"
        resp = self.__request(uri, "put",
            response_type=MediaType.ORIGIN_RESPONSE)
        return resp

class HostPool:
    """
    host管理器，后续继承接口请求，并对请求错误做标记

    一组ip每次获取下一个，如果网络请求失败会标记次数，次数到了阈值会临时移除，并在一点时间后重新加入
    """
    idx = 0
    hosts: List[HostClient] = []

    # 失败次数，到达次数后会将ip移动到失败数据中，并且标记失败时间
    _fallback_count = 5
    # 失败后将host放回的间隔
    _fallback_init_sec = 600
    # host和上次失败时间映射
    _fallback_time_dict = {}
    # host和失败次数映射
    _fallback_count_dict = {}
    # 失败恢复时间系数
    _fallback_factor_dict = {}

    def __init__(self, host="127.0.0.1:8848"):
        ar = [str.strip(x) for x in host.split(",")]
        hosts = [x if ":" in x else f"{x}:8848" for x in ar]
        self.hosts = [HostClient(host=x, trys=self._fallback_count)
            for x in hosts]
        self.size = len(self.hosts)

    def borrow(self) -> HostClient:
        """
        获取一个客户端，客户端包含对nacos的各种操作，每个client都是无状态的服务，绑定一个固定的host.
        如果由于连接超时等多次链接失败会将该host的client暂时移除，无法获取，在等待一段时候后恢复可用.

        Returns:
        获取一个连接对象

        Examples:
        HostPool(host="127.0.0.1:8848").borrow()
        """
        now = dt.time()
        hosts = []
        for h in self.hosts:
            if h.is_valid and h.trys <= 0:
                h.is_valid = False
                self._fallback_time_dict[h.host] = now
            elif not h.is_valid:
                fac = dict.get(self._fallback_factor_dict, h.host, 1)
                if (now - dict.get(self._fallback_time_dict, h.host, 0)
                    ) > fac * self._fallback_init_sec :
                    h.is_valid = True
                    # 恢复的时候把时间系数加一
                    self._fallback_factor_dict[h.host] = dict.get(
                        self._fallback_factor_dict, h.host, 1) + 1
            else:
                hosts.append(h)

        size = len(hosts)
        if not size:
            raise Exception("[Nacos] 没有可用的服务地址")
        i = min(self.idx, size - 1)
        logger.debug("[Nacos] Borrow Client : Total [%s], Now [%s]",
            size, self.idx)
        try:
            return hosts[i]
        finally:
            self.idx = (i+1) % size

def get_host_ip():
    """
    获取本机ip

    Returns:
    本机ip地址
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip

def get_access_token(host:HostPool, username="", password="") -> Any:
    # pylint: disable=line-too-long
    """
    获取nacos认证token

    Args:
      host: 连接池
      username: 用户名
      password: 密码

    Returns:
      如果成功：{"accessToken": "eyJhbGciOiJIUzI1NiJ9.eyJzdWIiOiJuYWNvcyIsImV4cCI6MTY0Njg5MzYzMH0.5ZeE0htGExkKrnp99sVKetue02obkx7OsJ85rHFMZIY", "tokenTtl": 18000, "globalAdmin": True, "username": "nacos"}
      如果失败：None
    """
    data = None
    try:
        while True:
            h = host.borrow()
            if not h or not h.is_health:
                return data
            data = h.login(username=username, password=password)
            if data:
                return data
    except RequestConnectionError as e:
        logger.error("连接超时 %s", str(e))
    except Exception as e:
        logger.exception(e, exc_info=True)
    return data



"""
build for publish script.
"""
import re

import setuptools

module_dir = "nacos_starter"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open(f"src/{module_dir}/__init__.py", encoding="utf-8") as fh:
    lines = fh.readlines()
    vl = [x for x in lines if "__version__" in x][0]
    version = re.sub(r"[^\d.]", "", str.strip(vl.split("=")[1]))
    namel = [x for x in lines if "name" in x][0]
    name = re.sub(r"[^\w\-_]", "", str.strip(namel.split("=")[1]))
# read dependencies from requirements.txt
# pylint: disable=consider-using-with
install_requires=[line.strip() for line in open("requirements.txt", encoding="utf-8").readlines()]

setuptools.setup(
    name=name,
    version=version,
    author="chizongyang",
    author_email="chizongyang@aliyun.com",
    description="Nacos starter for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/asdw741111/nacos-py-starter",
    project_urls={
        "Bug Tracker": "https://github.com/asdw741111/nacos-py-starter/issues",
        "Source Code": "https://github.com/asdw741111/nacos-py-starter",
    },
    include_package_data=True,
    package_dir={"": "src"},
    # 自动查找模块
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=install_requires,
    python_requires=">=3.6",
)



"""
decorator for service client, same as feign.
"""
import ctypes
import inspect
import threading
import time

from .util import logger


def default_fall_back_func(*args, **kwargs):
    return ("function fallback " +
        "  function info: " + str(args) + " " + str(kwargs))


def default_timeout_fallback_func(ex, *args, **kwargs):
    # raise ex
    return (
        "function time out "
        + str(ex)
        + "  function info: "
        + str(args)
        + " "
        + str(kwargs)
    )


def default_except_fallback_func(ex, *args, **kwargs):
    # raise ex
    logger.error("fallback %s fun info: %s %s", ex, args, kwargs)
    return (
        "function except "
        + str(ex)
        + "  function info: "
        + str(args)
        + " "
        + str(kwargs)
    )


class DegradeRule(object):
    """请求接口装饰器，用于配置一个要请求的服务接口"""

    def __init__(
        self,
        fallback_func=default_fall_back_func,
        timeout_fallback_func=default_timeout_fallback_func,
        except_fallback_func=default_except_fallback_func,
    ):
        self._func_time = {}
        self.fallback_func = fallback_func  # 熔断回调函数
        self.timeout_fallback_func = timeout_fallback_func  # 未触发熔断时 超时响应的回调函数
        self.except_fallback_func = except_fallback_func  # 未触发熔断时 异常响应的回调函数
        self._func_fuse = {}  # 存放熔断信息
        self.__kwargs = None
        self.__args = None

    def __async_raise(self, tid, exctype=SystemExit):
        """raises the exception, performs cleanup if needed"""
        tid = ctypes.c_long(tid)
        if not inspect.isclass(exctype):
            exctype = type(exctype)
        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid,
            ctypes.py_object(exctype))
        if res == 0:
            raise ValueError("invalid thread id")
        elif res != 1:
            # """if it returns a number greater than one, you"re in trouble,
            # and you should call it again with exc=NULL to revert the effect"""
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
            raise SystemError("PyThreadState_SetAsyncExc failed")

    def __go_func(self, tid, fn, fuse_status):
        try:
            objs = fn(*self.__args, **self.__kwargs)
            self._func_time[fn.__name__ + str(tid)] = objs
        except Exception as ex:
            fc = self.except_fallback_func(
                ex, fn.__name__, *self.__args, **self.__kwargs
            )
            self._func_time[fn.__name__ + str(tid)] = fc
            ec = self._func_fuse[fn]["error"] + 1
            if fuse_status:
                self._func_fuse[fn]["error"] = ec

    def __is_fuse(self,
        fn, fuse_status, except_percent, time_windows, time_count):
        if fuse_status is False:
            return False
        try:
            self._func_fuse[fn]
        except KeyError:
            self._func_fuse[fn] = {}
            self._func_fuse[fn]["count"] = 0
            self._func_fuse[fn]["error"] = 0
            self._func_fuse[fn]["time_count"] = time.time()
            self._func_fuse[fn]["time_windows"] = 0
        x = time.time() - self._func_fuse[fn]["time_count"]
        if x > time_count:
            time_windows_now = self._func_fuse[fn]["time_windows"]
            if time_windows_now != 0:  # 不等于0则认为触发熔断
                x1 = time.time() - time_windows_now  # 判断是否还在熔断时间窗口期
                if x1 <= time_windows:
                    return True
                else:  # 结束熔断窗口期 重置统计异常窗口期
                    self._func_fuse[fn]["time_count"] = time.time()
                    self._func_fuse[fn]["time_windows"] = 0
                    return False
            else:  # 等于0则判断当前行为是否熔断
                count = self._func_fuse[fn]["count"]
                error = self._func_fuse[fn]["error"]
                self._func_fuse[fn]["count"] = 0  # 取出来以后立刻重置
                self._func_fuse[fn]["error"] = 0
                if count != 0:
                    e = error / count
                    if e > except_percent:
                        self._func_fuse[fn]["time_windows"] = time.time()
                        return True
                    else:  # 未达到触发条件，继续重置统计异常窗口期
                        self._func_fuse[fn]["time_count"] = time.time()
                        return False
                else:  # 没有执行过函数 重置窗口期继续统计
                    self._func_fuse[fn]["time_count"] = time.time()
                    return False
        else:
            return False

    def hystrix(
        self,
        timeout=6,
        fuse_status=False,
        except_percent=0.5,
        time_windows=5,
        time_count=2,
    ):
        """熔断控制

        Args:
            timeout (int, optional): 超时时间. Defaults to 6.
            fuse_status (bool, optional): _description_. Defaults to False.
            except_percent (float, optional): 熔断阈值, 时间窗口内失败请求数占总请求数比例, 0~1之间. Defaults to 0.5.
            time_windows (int, optional): 时间窗口, 单位秒. Defaults to 5.
            time_count (int, optional): 时间窗口内用于判断熔断的总请求数, 达到该值才会判断是否需要熔断. Defaults to 2.
        """
        def get_f(fn):
            def get_args(*args, **kwargs):
                self.__kwargs = kwargs
                self.__args = args
                # 判断是否需要熔断，默认不熔断
                if self.__is_fuse(
                    fn, fuse_status, except_percent, time_windows, time_count
                ):
                    return self.fallback_func(fn.__name__, *args, **kwargs)
                tid = time.time()
                func_thread = threading.Thread(
                    target=self.__go_func, args=(tid, fn, fuse_status), daemon=True
                )
                func_thread.start()
                func_thread.join(timeout)
                try:
                    self.__async_raise(func_thread.ident)
                    tff = self.timeout_fallback_func(
                        TimeoutError, fn.__name__, *args, **kwargs
                    )
                    self._func_time[fn.__name__ + str(tid)] = tff
                    if fuse_status:
                        ffe = self._func_fuse[fn]["error"] + 1
                        self._func_fuse[fn]["error"] = ffe
                except ValueError:
                    pass
                if fuse_status:
                    ffc = self._func_fuse[fn]["count"] + 1
                    self._func_fuse[fn]["count"] = ffc
                return self._func_time[fn.__name__ + str(tid)]

            get_args.__name__ = fn.__name__
            return get_args

        return get_f


class FuncFlowControl:
    """请求控制，对请求失败进行处理"""

    def __init__(self, fall_back_func=default_fall_back_func):
        self.fall_back_func = fall_back_func
        self._flow_control_msg = {}
        # self.flowControlMsg = self.__flowControlMsg
        self._func_return = None

    def flow_control(self, time_windows, max_count):
        """流量控制

        Args:
            time_windows (int): 时间窗口, 单位秒, 表示从当前时间往前推多少秒用于计算最大请求数
            max_count (int): 时间窗口内最大请求数, 如果超出则执行fall_back_func
        """
        def get_f(fn):
            try:
                self._flow_control_msg[fn]
            except KeyError:
                self._flow_control_msg[fn] = 0
                self._flow_control_msg[str(fn) + "time"] = time.time()

            def get_args(*args, **kwargs):
                x = time.time() - self._flow_control_msg[str(fn) + "time"]
                if x < time_windows:
                    if self._flow_control_msg[fn] >= max_count:
                        return self.fall_back_func(fn.__name__, *args, **kwargs)
                else:
                    self._flow_control_msg[fn] = 0
                    self._flow_control_msg[str(fn) + "time"] = time.time()
                self._flow_control_msg[fn] = self._flow_control_msg[fn] + 1
                self._func_return = fn(*args, **kwargs)
                self._flow_control_msg[fn] = self._flow_control_msg[fn] - 1
                return self._func_return

            get_args.__name__ = fn.__name__
            return get_args

        return get_f



"""
Constants config file.
"""
# pylint: disable=line-too-long
LOG_FORMAT = (
    "%(asctime)s -%(name)s- %(threadName)s-%(thread)d - %(levelname)s - %(message)s"
)
DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"
TIME_OUT = 10
BEAT_TIME = 10
REGISTER_DICT_KEY = "healthy"
